"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division
import numpy as np
import cv2
from PIL import Image

from .transform import change_transform_origin


def read_image_bgr(path):
    """ Read an image in BGR format.

    Args
        path: Path to the image.
    """
    image = np.asarray(Image.open(path).convert('RGB'))
    return image[:, :, ::-1].copy()


def preprocess_image(x, mode='caffe'):
    """ Preprocess an image by subtracting the ImageNet mean.

    Args
        x: np.array of shape (None, None, 3) or (3, None, None).
        mode: One of "caffe" or "tf".
            - caffe: will zero-center each color channel with
                respect to the ImageNet dataset, without scaling.
            - tf: will scale pixels between -1 and 1, sample-wise.

    Returns
        The input with the ImageNet mean subtracted.
    """
    # mostly identical to "https://github.com/keras-team/keras-applications/blob/master/keras_applications/imagenet_utils.py"
    # except for converting RGB -> BGR since we assume BGR already

    # covert always to float32 to keep compatibility with opencv
    x = x.astype(np.float32)

    if mode == 'tf':
        #print("Z-score")
        x -= np.mean(x, keepdims=True)
        x /= (np.std(x, keepdims=True) + 1e-6)
        #x /= 127.5
        #x -= 1.
    elif mode == 'caffe':
        x[..., 0] -= 103.939
        x[..., 1] -= 116.779
        x[..., 2] -= 123.68

    return x


def adjust_transform_for_image(transform, image, relative_translation):
    """ Adjust a transformation for a specific image.

    The translation of the matrix will be scaled with the size of the image.
    The linear part of the transformation will adjusted so that the origin of the transformation will be at the center of the image.
    """
    height, width, channels = image.shape

    result = transform

    # Scale the translation with the image size if specified.
    if relative_translation:
        result[0:2, 2] *= [width, height]

    # Move the origin of transformation.
    result = change_transform_origin(transform, (0.5 * width, 0.5 * height))

    return result


class TransformParameters:
    """ Struct holding parameters determining how to apply a transformation to an image.

    Args
        fill_mode:             One of: 'constant', 'nearest', 'reflect', 'wrap'
        interpolation:         One of: 'nearest', 'linear', 'cubic', 'area', 'lanczos4'
        cval:                  Fill value to use with fill_mode='constant'
        relative_translation:  If true (the default), interpret translation as a factor of the image size.
                               If false, interpret it as absolute pixels.
    """
    def __init__(
        self,
        fill_mode            = 'nearest',
        interpolation        = 'linear',
        cval                 = 0,
        relative_translation = True,
    ):
        self.fill_mode            = fill_mode
        self.cval                 = cval
        self.interpolation        = interpolation
        self.relative_translation = relative_translation

    def cvBorderMode(self):
        if self.fill_mode == 'constant':
            return cv2.BORDER_CONSTANT
        if self.fill_mode == 'nearest':
            return cv2.BORDER_REPLICATE
        if self.fill_mode == 'reflect':
            return cv2.BORDER_REFLECT_101
        if self.fill_mode == 'wrap':
            return cv2.BORDER_WRAP

    def cvInterpolation(self):
        if self.interpolation == 'nearest':
            return cv2.INTER_NEAREST
        if self.interpolation == 'linear':
            return cv2.INTER_LINEAR
        if self.interpolation == 'cubic':
            return cv2.INTER_CUBIC
        if self.interpolation == 'area':
            return cv2.INTER_AREA
        if self.interpolation == 'lanczos4':
            return cv2.INTER_LANCZOS4


def apply_transform(matrix, image, params):
    """
    Apply a transformation to an image.

    The origin of transformation is at the top left corner of the image.

    The matrix is interpreted such that a point (x, y) on the original image is moved to transform * (x, y) in the generated image.
    Mathematically speaking, that means that the matrix is a transformation from the transformed image space to the original image space.

    Args
      matrix: A homogeneous 3 by 3 matrix holding representing the transformation to apply.
      image:  The image to transform.
      params: The transform parameters (see TransformParameters)
    """
    output = cv2.warpAffine(
        image,
        matrix[:2, :],
        dsize       = (image.shape[1], image.shape[0]),
        flags       = params.cvInterpolation(),
        borderMode  = params.cvBorderMode(),
        borderValue = params.cval,
    )
    return output


def compute_resize_scale(image_shape, min_side=800, max_side=1333):
    """ Compute an image scale such that the image size is constrained to min_side and max_side.

    Args
        min_side: The image's min side will be equal to min_side after resizing.
        max_side: If after resizing the image's max side is above max_side, resize until the max side is equal to max_side.

    Returns
        A resizing scale.
    """
    (rows, cols, _) = image_shape

    smallest_side = min(rows, cols)

    # rescale the image so the smallest side is min_side
    scale = min_side / smallest_side

    # check if the largest side is now greater than max_side, which can happen
    # when images have a large aspect ratio
    largest_side = max(rows, cols)
    if largest_side * scale > max_side:
        scale = max_side / largest_side

    return scale


def resize_image(img, min_side=800, max_side=1333):
    """ Resize an image such that the size is constrained to min_side and max_side.

    Args
        min_side: The image's min side will be equal to min_side after resizing.
        max_side: If after resizing the image's max side is above max_side, resize until the max side is equal to max_side.

    Returns
        A resized image.
    """
    # compute scale to resize the image
    scale = compute_resize_scale(img.shape, min_side=min_side, max_side=max_side)

    # resize the image with the computed scale
    img = cv2.resize(img, None, fx=scale, fy=scale)

    return img, scale

def _uniform(val_range):
    """
    Uniformly sample from the given range

    Args
        val_range: A pair of lower and upper bound
    """
    return np.random.uniform(val_range[0],val_range[1])

def _check_range(val_range, min_val=None, max_val=None):
    """
    Check whether the range is a valid range

    Args
        val_range: A pair of lower and upper bound
        min_val: minimal lower bound
        max_val: maximum upper bound
    """
    if val_range[0] > val_range[1]:
        raise ValueError('interval lower bound > upper bound')
    if min_val is not None and val_range[0] < min_val:
        raise ValueError('invalid interval lower bound')
    if max_val is not None and val_range[1] > max_val:
        raise ValueError('invalid interval upper bound')

def _clip(image):
    """
    Clip and convert an image to np.uint8

    Args
        image: Image to clip
    """
    return np.clip(image, 0, 255).astype(np.uint8)

class VisualEffect:
    """ Struct holding parameters for applying a visual effect.

    Args
        contrast_factor:   contrast factor between 0 and 3
        brightness_delta:  brightness offset between -1 and 1
        hue_delta:         hue offset -1 and 1
        saturation_factor: nonnegative saturation factor
    """

    def __init__(
        self,
        contrast_factor,
        brightness_delta,
        hue_delta,
        saturation_factor,
    ):
        self.contrast_factor = contrast_factor
        self.brightness_delta = brightness_delta
        self.hue_delta = hue_delta
        self.saturation_factor = saturation_factor

def random_visual_effect_generator(
    contrast_range=(0.5,2.),
    brightness_range=(-.25,.25),
    hue_range=(-0.1,0.1),
    saturation_range=(0.8,1.2),
):
    """
    Generate visual effect parameters sampled from the given intervals.

    Args
        contrast_factor:   interval between 0 and 3
        brightness_delta:  interval between -1 and 1
        hue_delta:         interval between -1 and 1
        saturation_factor: interval with the lower bound 0
    """
    _check_range(contrast_range, 0)
    _check_range(brightness_range, -1, 1)
    _check_range(hue_range, -1, 1)
    _check_range(saturation_range, 0)

    def _generate():
        while True:
            yield VisualEffect(
                contrast_factor=_uniform(contrast_range),
                brightness_delta=_uniform(brightness_range),
                hue_delta=_uniform(hue_range),
                saturation_factor=_uniform(saturation_range),
            )
    return _generate()

def adjust_contrast(image, factor):
    """
    Adjust contrast of an image

    Args
        image: image to adjust
        factor: contrast factor
    """
    mean = image.mean(axis=0).mean(axis=0)
    return _clip((image - mean) * factor + mean)

def adjust_brightness(image, delta):
    """
    Adjust brightness of an image

    Args
        image: image to adjust
        delta: brightness offset in range [-1,1]
    """
    return _clip(image + delta * 255)

def adjust_hue(image, delta):
    """
    Adjust hue of an image

    Args
        image: image to adjust
        delta: hue offset in range [-1,1]
    """
    image[...,0] = np.mod(image[...,0] + delta * 180, 180)
    return image

def adjust_saturation(image, factor):
    """
    Adjust saturation of an image

    Args
        image: image to adjust
        factor: saturation factor
    """
    image[...,1] = np.clip(image[...,1] * factor, 0 , 255)
    return image

def apply_visual_effect(effect, image):
    """
    Apply a visual effect on the image

    Args
        effect: visual effect parameters
        image: Image to adjust
    """

    if effect.contrast_factor:
        image = adjust_contrast(image, effect.contrast_factor)
    if effect.brightness_delta:
        image = adjust_brightness(image, effect.brightness_delta)

    if effect.hue_delta or effect.saturation_factor:

        image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

        if effect.hue_delta:
            image = adjust_hue(image, effect.hue_delta)
        if effect.saturation_factor:
            image = adjust_saturation(image, effect.saturation_factor)

        image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)

    return image